from optparse import OptionParser
from utilities.overlap_coord import OverlapCoord
from utilities.tools import reverse_complement
import os, re

fields = ['bin', 'chrom', 'chromStart', 'chromEnd', 'name', 'score', 'strand',
          'refNCBI', 'refUCSC', 'observed', 'molType', 'class', 'valid', 'avHet',
          'avHetSE', 'func', 'locType', 'weight']

def parse_line(line):
    cols = line.rstrip('\n').split('\t')
    
    data = {}
    for i in range(len(cols)):
	if i == len(fields):
	    break
	
        data[fields[i]] = cols[i]
        
    if data['class'] == 'single':
        data['type'] = 'snv'
    elif data['class'] == 'deletion':
        data['type'] = 'del'
    elif data['class'] == 'insertion':
        data['type'] = 'ins'
    else:
        data['type'] = 'NA'
        
    if data['type'] == 'ins':
	data['start'] = int(data['chromStart'])
    else:
	data['start'] = int(data['chromStart']) + 1
    data['end'] = int(data['chromEnd'])
    data['allele'] = {}
    data['size'] = 0
    for a in data['observed'].split('/'):
        if a != '-':
	    if data['strand'] == '+':
		data['allele'][a.lower()] = True
	    elif data['strand'] == '-':
		data['allele'][reverse_complement(a).lower()] = True
		
            if data['size'] == 0:
                data['size'] = len(a)
        
    if data['observed'] == 'lengthTooLong':
	data = {}
	
    return data

def index(input, output):
    indices = {}
    data_file = os.path.abspath(input)
    line_num = 1
    for line in open(input, 'r'):
        cols = line.rstrip().split("\t")

        start = int(int(cols[2])/1000)
        end = int(int(cols[3])/1000)
        target = cols[1]

        if not 'chr' in target:
            target = 'chr' + target

        for n in range(start,end+1):
            index = ':'.join((target,str(n)))
            value = str(line_num)

            if not indices.has_key(index):
                indices[index] = [value]
            else:
                indices[index].append(value)

        line_num += 1

    index_file = open(output, 'w')
    for index in sorted(indices.keys()):
        index_file.write(' '.join((index, ','.join(indices[index]))) + "\n")
        

def find_concordants(test, genome=None, refseq=None):
    concordants = []
    
    package_dir = "/".join(os.path.abspath(__file__).split("/")[:-3])
    genome_dir = package_dir + '/annotations/' + genome

    if not os.path.isdir(genome_dir):
	return concordants
    
    snp_file = genome_dir + '/snp/' + test['chrom'] + '-snp.txt'
    index_file = snp_file + '.idx'
    
    snp_overlap = OverlapCoord(snp_file, index_file)
    snp_overlap.extract_index()
    
    knowns = snp_overlap.overlap(test['chrom'], test['start'], test['end'], parse_line=parse_line)
    
    for known in [k for k in knowns if k['type'].lower() == test['type'].lower() and k['size'] == test['size']]:
        if is_concordant(known, test, refseq):
	    concordants.append(known['name'])
	    
    if test['type'] == 'ins' or test['type'] == 'del':
	for known in [k for k in knowns if k['class'].lower() == 'in-del']:	    
	    if is_part_of_indel(known, test, refseq):
		concordants.append(known['name'])
		
    snp_overlap.finish()
    return concordants
        
def is_concordant(known, test, refseq=None):
    concordant = False
    
    if test['type'] == known['type']:
        if test['chrom'] == known['chrom']:
            if test['start'] == known['start'] and test['end'] == known['end']:
                if test['type'] == 'del':
                    concordant = True
                elif known['allele'].has_key(test['allele']) and test['allele'] != 'NA':
                    concordant = True
                    
    if not concordant and (test['type'] == 'ins' or test['type'] == 'del') and test['size'] == known['size'] and refseq:
        concordant = check_flanking(known, test, refseq)
            
    return concordant

def is_part_of_indel(known, test, refseq=None):
    """
    in-del is a non-zero-length ref sequence replaced by a non-zero-length allele
    this is split into a del and an ins from TA
    so this checks whether the TA events is either
    a) a deletion that matches the in-del deletion
    b) an insertion that matches one of the in-del allele
    """
    
    concordant = False
        
    if test['type'] == 'del':	
	known_start = int(known['chromStart']) + 1
	
	if test['size'] == len(known['refUCSC']):
	    if test['start'] == known_start and test['end'] == known['end']:
		concordant = True
	    else:
		concordant = check_flanking(known, test, refseq)
		
    elif test['type'] == 'ins':	
	known_start = int(known['chromStart']) 
	
	if test['start'] == known_start and known['allele'].has_key(test['allele']) and test['allele'] != 'NA':
	    concordant = True
	else:
	    concordant = check_flanking(known, test, refseq)		
	
    return concordant
                    
def check_flanking(known, test, refseq, flank_size=1000):
    same = False
        
    start = min(known['start'], test['start']) - flank_size
    end = max(known['end'], test['end']) + flank_size
    	    
    if test['type'] == 'ins':
	test_seq = refseq.GetSequence(test['chrom'], start, test['start']) + test['allele'] + refseq.GetSequence(test['chrom'], test['start'] + 1, end)
	    
	for allele in known['allele'].keys():
	    #known_seq = chrom_seq[start-1:known['start']] + allele + chrom_seq[known['start']:end]
	    known_seq = refseq.GetSequence(test['chrom'], start, known['start']) + allele + refseq.GetSequence(test['chrom'], known['start'] + 1, end)
    
	    if known_seq.lower() == test_seq.lower():
		same = True
		break
		    
    elif test['type'] == 'del':
	test_seq = refseq.GetSequence(test['chrom'], start, test['start'] - 1) + refseq.GetSequence(test['chrom'], test['end'] + 1, end)
	known_seq = refseq.GetSequence(test['chrom'], start, known['start'] - 1) + refseq.GetSequence(test['chrom'], known['end'] + 1, end)
				    
	if known_seq.lower() == test_seq.lower():
	    same = True
		    
    return same

if __name__ == '__main__':
    usage = "Usage: %prog annotation-file"
    parser = OptionParser(usage=usage)
    parser.add_option("-i", "--index", dest="index", help="index output file")

    (options, args) = parser.parse_args()
    
    if options.index:
        index(args[0], options.index)
