from annotations import ensembl, knownGene, refGene, aceview, ensg
from utilities import track, assembly
from utilities import tools, intspan
from optparse import OptionParser
import ConfigParser, os, sys, re, glob
import event, event_support
import pprint
import linecache
from utilities.bam import BAM
from utilities.align_parsers import blat

VERSION = '1.3.2'

PACKAGE_DIR = "/".join(os.path.abspath(sys.argv[0]).split("/")[:-2])
CONFIG_FILE = PACKAGE_DIR + "/configs/model_matcher.cfg"
pp = pprint.PrettyPrinter(indent=0)

class MatchResult:

    def __init__(self, align_id, txt_id, num_blocks, num_exons, model=None, txt=None):
        self.align_id = align_id
        self.txt_id = txt_id
        self.num_blocks = num_blocks
        self.num_exons = num_exons
        self.txt = txt
        self.model = model
        self.matched_blocks = []
        self.matched_exons = []
        self.events = []
        self.coverage = 0.0
        self.model_weight = None
        self.first = self.last = self.middle = 0
        self.coord = ""
        

    def score(self):
        score = float(self.middle) + float(self.first) + float(self.last)

        for e in self.events:
            if e.type == "AS5" or e.type == "AS3":
                score += 0.5
            elif e.type == "AS53":
                score += 1.0
        
        return score

    def set_coverage(self, exon_bases, txt_len):
        self.coverage = float(exon_bases)/float(txt_len)

    @classmethod
    def compare_matches(cls, m1, m2):
        if m1.score() > m2.score():
            return -1
        elif m1.score() < m2.score():
            return 1
        elif m1.num_blocks == m1.num_exons and m2.num_blocks != m2.num_exons:
            return -1
        elif m1.num_blocks != m1.num_exons and m2.num_blocks == m2.num_exons:
            return 1
        elif int(m1.middle) > int(m2.middle):
            return -1
        elif int(m1.middle) < int(m2.middle):
            return 1
        else:
            if int(m1.first) or int(m1.last) or int(m2.first) or int(m2.last):
                if int(m1.first) + int(m1.last) > int(m2.first) + int(m2.last):
                    return -1
                elif int(m1.first) + int(m1.last) < int(m2.first) + int(m2.last):
                    return 1
                elif m1.events or m2.events:
                    if m1.events and not m2.events:
                        return 1
                    elif not m1.events and m2.events:
                        return -1
                    elif len(m1.events) < len(m2.events):
                        return -1
                    elif len(m2.events) < len(m1.events):
                        return 1
                    else:
                        if m1.model_weight > m2.model_weight:
                            return -1
                        elif m1.model_weight < m2.model_weight:
                            return 1
                        else:
                            if m1.txt.cds_length() > m2.txt.cds_length():
                                return -1
                            elif m1.txt.cds_length() < m2.txt.cds_length():
                                return 1
                            return 0
                else:
                    if m1.model_weight > m2.model_weight:
                        return -1
                    elif m1.model_weight < m2.model_weight:
                        return 1
                    else:
                        if m1.txt.cds_length() > m2.txt.cds_length():
                            return -1
                        elif m1.txt.cds_length() < m2.txt.cds_length():
                            return 1
                        return 0
                    
            else:
                if m1.events and not m2.events:
                    return 1
                elif not m1.events and m2.events:
                    return -1
                elif len(m1.events) < len(m2.events):
                    return -1
                elif len(m2.events) < len(m1.events):
                    return 1
                else:
                    if m1.model_weight > m2.model_weight:
                        return -1
                    elif m1.model_weight < m2.model_weight:
                        return 1
                    else:
			if m1.txt.cds_length() > m2.txt.cds_length():
			    return -1
			elif m1.txt.cds_length() < m2.txt.cds_length():
			    return 1
			return 0 
                

    def full_match(self):
        if self.num_blocks == len(self.matched_blocks) and self.num_exons == len(self.matched_blocks) and int(self.middle) == (self.num_exons-2) * 2 and int(self.first) > 0 and int(self.last) > 0 and len(self.events) == 0:
            return True
        else:
            return False

    def report(self, details=False, out=None):
        report = ""
        
        if not self.matched_blocks or self.matched_blocks == 0:
            report = "%s no_match total_blocks=%d\n" % (self.align_id, self.num_blocks)
        else:
            if self.full_match():
                status = "full_match"
            else:
                status = "partial_match"
		
	    coding_type = self.txt.coding_type()
		
            report = "%s matches %s %s model:%s(wt:%d) in %d blocks total_blocks=%d total_exons=%d %s coord:%s score:%.1f events:%d coverage:%s\n" % (self.align_id, self.txt_id, coding_type, self.model, self.model_weight, len(self.matched_blocks), self.num_blocks, self.num_exons, status, self.coord, self.score(), len(self.events), "%.3f" % (self.coverage))
            if details:
                for i in range(len(self.matched_blocks)):
                    report += "block %d = exon %d\n" % (self.matched_blocks[i], self.matched_exons[i])

        if out:
            out.write(report)
        else:
            return report


class ModelMatcher:

    def __init__(self, aligns=[], genome=None, model_order=[], debug=False, assembly=None, outdir=None, contigs_bam=None, genome_bam=None, lib=None, min_spanning_reads=None, max_coverage_diff=None, refseq=None):
        self.aligns = aligns
        self.genome = genome
        self.model_order = model_order
        self.log_file = outdir + "/log.txt"
        self.debug = debug
        self.outdir = outdir
        
        self.min_spanning_reads = min_spanning_reads
        self.max_coverage_diff = max_coverage_diff

        if self.log_file:
            self.log = open(self.log_file, 'w')
        else:
            self.log = None

        if self.genome:
            self.get_models()

        self.assembly = assembly

        #for event support
        self.contigs_bam = self.genome_bam = self.lib = None
        if contigs_bam:
            self.contigs_bam = BAM(contigs_bam, min_mapq=0)
            self.lib = lib
        if genome_bam:
            self.use_genome = True
            self.genome_bam = BAM(genome_bam, min_mapq=1)
        else:
            self.use_genome = False

        self.refseq = refseq
	self.add_contig_reads = False
	self.coverage_model = None

    def get_models(self):
        "get gene model info from config file info"

        self.annots = {}
        self.indices = {}
        
        self.config = ConfigParser.ConfigParser()
        self.config.read(CONFIG_FILE)
        if self.genome in self.config.sections():
            for field in self.config.options(self.genome):
                value = self.config.get(self.genome, field)

                if field == "order" and not self.model_order:
                    self.model_order = value.split(',')
                else:
                    self.annots[field] = value

            if self.annots:
                annot_path = PACKAGE_DIR + "/annotations/" + self.genome
                for m,f in self.annots.iteritems():
                    full_path = annot_path + "/" + f
                    if self.log:
                        self.log.write(m + ":" + full_path + "\n")
                    if os.path.exists(full_path):
                        self.annots[m] = full_path
                        self.indices[m] = full_path.replace(".txt", ".idx")

        self.splice_file = PACKAGE_DIR + "/annotations/" + self.genome + "/splice_motifs.txt"
	self.splice_motifs = tools.get_splice_motifs(self.splice_file)

        if self.log:
            self.log.write("model order: %s\n" % ','.join(self.model_order))
	    

    def match_models(self):
        sys.stdout.write("extracting transcript indices...")
        models = {}
        for i in range(len(self.model_order)):
            model = self.model_order[i]
            model_file = self.annots[model]

            if not self.annots.has_key(model) or not os.path.exists(model_file):
                if self.log:
                    self.log.write("doesn't have model: %s %s\n" % (model, model_file))
                continue

            models[model] = {}
            models[model]['source'] = self.annots[model]
            models[model]['weight'] = len(self.model_order) - i

            if os.path.exists(self.indices[model]):
                models[model]['index'] = self.extract_index(self.indices[model])
        sys.stdout.write("done\n")
            
        tobjs = {}
        overlapping_txts = []
        
        events = []

        contig2txts = {}

        mapping_outfile = self.outdir + "/mapping.txt"
        if os.path.exists(mapping_outfile):
            os.remove(mapping_outfile)
        mapping_out = open(mapping_outfile, 'w')
        
        count = 1
        sys.stdout.write("starts matching %d alignments...\n" % len(self.aligns))
	
        for align in self.aligns:
	    if self.refseq and align.contig and align.contig.sequence:
		align.correct_blocks(self.splice_motifs, self.refseq, align.contig.sequence)
		align.merge_blocks()
	    
            #find out overlapping txts
            del overlapping_txts[:]
            for model in self.model_order:
                if not tobjs.has_key(model):
                    tobjs[model] = {}
                if not contig2txts.has_key(model):
                    contig2txts[model] = {}
                
                index = models[model]['index']
                source = models[model]['source']
                txts = self.overlap_txts(align.target, align.blocks[0][0], align.blocks[-1][1], index, source, model)

                for txt in txts:
		    if not txt:
			continue
                    txt.model = model
                    txt.weight = models[model]['weight']
                    if not tobjs[model].has_key(txt.name):
                        tobjs[model][txt.name] = txt

                    overlapping_txts.append(tobjs[model][txt.name])

            matches = self.compare_model(align, overlapping_txts)
            matches.sort(MatchResult.compare_matches)
            best_is_full = matches[0].full_match()

            #output matches plus storing info for coverage
            for match in matches:
                if not match.matched_blocks or match.matched_blocks == 0:
                    continue
                
                txt = match.txt
                
                #don't output partial match if best match is a full match
                if best_is_full:
                    if match.full_match():
                        match.report(out=mapping_out)

                        if not contig2txts[txt.model].has_key(txt.name):
                            contig2txts[txt.model][txt.name] = {}
                        contig2txts[txt.model][txt.name][align.query] = align, match.coverage

                else:
                    match.report(out=mapping_out)

                    if not contig2txts[txt.model].has_key(txt.name):
                        contig2txts[txt.model][txt.name] = {}
                    contig2txts[txt.model][txt.name][align.query] = align, match.coverage
                        
            #capture events of best match
            if matches[0].events:
                if len(matches[0].matched_blocks) > 0:
                    events.extend(matches[0].events)
                else:
                    events.append(matches[0].events[0])
                remove_start = 1
            else:
                remove_start = 0

            #clean-up memory usage
            for i in range(remove_start, len(matches)):
                del matches[i].events[:]
            del matches[remove_start:]

            count += 1
            if count % 1000 == 0:
                sys.stderr.write("\rfinished %d alignments\n" % count)

        #get sequence for orf
        if self.assembly:
            contigs_with_events = dict((e.align.contig.num, e.align.contig) for e in events)
            self.assembly.get_seqs(contigs_with_events.values())

        event.Event.splice_motifs = self.splice_motifs
        good_events = [e for e in events if e.novel and not e.artefact]
        [e.details(self.refseq) for e in good_events]

        #screen
        if self.genome_bam:
            for e in good_events:
                event_support.find_support_genome(e, self.genome_bam)

        if self.contigs_bam:
            reads_dir = self.outdir + "/events_reads"
            if os.path.isdir(reads_dir):
                os.system("rm -Rf " + reads_dir)
            os.mkdir(reads_dir)
                
            for e in good_events:
                event_support.find_support_contig(e, self.contigs_bam, self.lib, outdir=reads_dir)
		
	    if self.add_contig_reads:
		self.add_contig_support(good_events)
		
	    for e in good_events:
                event_support.set_read_support(e, self.use_genome, min_spanning_reads=self.min_spanning_reads, max_coverage_diff=self.max_coverage_diff)
                event_support.screen(e)

        #output events
        events_outfile = self.outdir + "/events.txt"
        events_bedfile = self.outdir + "/events.bed"
        events_summary = self.outdir + "/events_summary.txt"
        grouped_events = self.group_events(events)
        self.output_events(grouped_events, events_outfile, events_summary, events_bedfile)

        events_passed = [e for e in good_events if e.filter == 'passed']
        events_outfile = self.outdir + "/events_filtered.txt"
        events_bedfile = self.outdir + "/events_filtered.bed"
        events_summary = self.outdir + "/events_filtered_summary.txt"
        grouped_events = self.group_events(events_passed)
        self.output_events(grouped_events, events_outfile, events_summary, events_bedfile)
        
        coverage_outfile = self.outdir + "/coverage.txt"
        self.output_coverage(contig2txts, tobjs, coverage_outfile)


    def add_contig_support(self, all_events):
	grouped_events = self.group_events(all_events)
	print grouped_events
	
	for event_type in grouped_events.keys():
	    #can not handle novel_transcript as number of blocks varies
	    if event_type == 'novel_transcript':
		continue
	    
	    for coord, events in grouped_events[event_type].iteritems():
		if len(events) > 1:
		    if event_type in ('AS53', 'novel_exon', 'novel_utr'):
			spanning_reads = []
			contig_coverage = []
		    else:
			spanning_reads = 0
			contig_coverage = 0
			
		    contig_neighbor = {'left':'na', 'right':'na'}
		
		    for e in events:
			print event_type, e.contig, e.spanning_reads, e.contig_coverage, e.contig_neighbor
			if e.spanning_reads != None:
			    if event_type in ('AS53', 'novel_exon', 'novel_utr'):
				if not spanning_reads:
				    for n in e.spanning_reads:
					spanning_reads.append(int(n))
				else:
				    for i in range(len(spanning_reads)):
					spanning_reads[i] += int(e.spanning_reads[i])
			    else:
				spanning_reads += int(e.spanning_reads)
				
			if e.contig_coverage != None:
			    if event_type in ('AS53', 'novel_exon', 'novel_utr'):
				if not contig_coverage:
				    for n in e.contig_coverage:
					contig_coverage.append(int(n))
				else:
				    for i in range(len(contig_coverage)):
					contig_coverage[i] += int(e.contig_coverage[i])
			    else:
				contig_coverage += int(e.contig_coverage)
		    
			if e.contig_neighbor:
			    for side in ('left', 'right'):
				if contig_neighbor[side] == 'na' and e.contig_neighbor[side] != 'na':
				    contig_neighbor[side] = 0
				if e.contig_neighbor[side] != 'na':
				    contig_neighbor[side] += e.contig_neighbor[side] 

		    print 'test', spanning_reads, contig_coverage, contig_neighbor
		    for e in events:
			e.spanning_reads = spanning_reads
			e.contig_coverage = contig_coverage
			e.contig_neighbor = contig_neighbor
			
    def extract_index(self, index_file):
        index = {}
        for line in open(index_file, 'r'):
            coord, lines = line.rstrip().split(" ")
            index[coord] = lines
        return index

    def overlap_txts(self, target, tstart, tend, index, source, model):
        tstart = int(tstart)
        tend = int(tend)
        tstart_index = ':'.join((target, str(int(tstart/1000))))
        tend_index = ':'.join((target, str(int(tend/1000))))

        txts = []
        txt_lines = {}
        if index.has_key(tstart_index):
            for line in index[tstart_index].split(','):
                txt_lines[line] = True
        if tend_index != tstart_index:
            int(int(tstart)/1000)
            for coord in range(int(tstart/1000)+1,int(tend/1000)+1):
                idx = ':'.join((target, str(coord)))
                if index.has_key(idx):
                    for line in index[idx].split(','):
                        txt_lines[line] = True

        for line_num in txt_lines.keys():
            line = linecache.getline(source, int(line_num))
            txt = {
                'e': ensembl.parse_line,
                'r': refGene.parse_line,
                'k': knownGene.parse_line,
                'a': aceview.parse_line,
                'x': ensg.parse_line
                }[model](line)
            txts.append(txt)

        return txts


    def as_artefacts(self, match):
        "remove matching AS5,AS3"
        
        novel_as5 = [e for e in match.events if e.novel and e.type == "AS5"]
        novel_as3 = [e for e in match.events if e.novel and e.type == "AS3"]

        as3_dict = {}
        for e in novel_as3:
            for block in e.align_blocks:
                as3_dict[int(block)] = e
                
        as5_dict = {}
        for e in novel_as5:
            for block in e.align_blocks:
                as5_dict[int(block)] = e
                
        for e in novel_as5:
            for block in e.align_blocks:
                if as3_dict.has_key(int(block)-1):
                    as3 = as3_dict[int(block)-1]
                                        
                    e.artefact = True
                    as3.artefact = True

                #depends on transcript orientation, as5 and as3 can mean left or right
                elif as3_dict.has_key(int(block)+1):
                    as3 = as3_dict[int(block)+1]
                                        
                    e.artefact = True
                    as3.artefact = True
                    

        novel_exons = [e for e in match.events if e.novel and e.type == "novel_exon"]

        #removes small novel exons <10 bp and associated with AS
        for e in novel_exons:
            for block in e.align_blocks:
                if intspan.cardinality(e.align.blocks[block-1]) < 10:
                    if block > 1 and block < len(e.align.blocks):
                        if as3_dict.has_key(block-1) and as3_dict[block-1].right:
                            e.artefact = True
                            as3_dict[block-1].artefact = True
                        if as5_dict.has_key(block-1) and as5_dict[block-1].right:
                            e.artefact = True
                            as5_dict[block-1].artefact = True
                        if as3_dict.has_key(block+1) and as3_dict[block+1].left:
                            e.artefact = True
                            as3_dict[block+1].artefact = True
                        if as5_dict.has_key(block+1) and as5_dict[block+1].left:
                            e.artefact = True
                            as5_dict[block+1].artefact = True
                        
                if e.artefact:
                    self.log.write(' '.join(("novel_exon artefact", e.contig, e.transcript, e.type, e.coordinate())) + "\n")

    def deletion_artefact(self, match, align):
        "remove events associated with gaps <= 20bp"

        #minimum intron size?
        min_intron_size = 20
        for e in match.events:
            if len(e.align_blocks) == 1:
                b = int(e.align_blocks[0])-1
                #check distance between previous block
                if b > 0 and int(align.blocks[b][0]) - int(align.blocks[b-1][1]) <= min_intron_size:
                    e.artefact = True
                    self.log.write(' '.join(("deletion artefact", e.contig, e.transcript, e.type, e.coordinate())) + "\n")
                #check distance between next block
                if b < len(align.blocks)-1 and int(align.blocks[b+1][0]) - int(align.blocks[b][1]) <= min_intron_size:
                    e.artefact = True
                    self.log.write(' '.join(("deletion artefact", e.contig, e.transcript, e.type, e.coordinate())) + "\n")
                    
            elif len(e.align_blocks) == 2:
                b1 = int(e.align_blocks[0])-1
                b2 = int(e.align_blocks[1])-1
                #check distance between blocks
                if int(align.blocks[b2][0]) - int(align.blocks[b1][1]) <= min_intron_size:
                    e.artefact = True
                    self.log.write(' '.join(("deletion artefact", e.contig, e.transcript, e.type, e.coordinate())) + "\n")

    def intron_artefact(self, match):
        "remove novel_intron event if it's not in exon"

        for e in match.events:
            if e.type == "novel_intron":
                if re.match('[\D]', str(e.txt.cdsStart)) or re.match('[\D]',  str(e.txt.cdsEnd)):
                    continue

                start, end = e.align_coords[0][0], e.align_coords[-1][1]
                #novel intron
                if len(e.align_coords) > 1:
                    start, end = e.align_coords[0][1], e.align_coords[-1][0]

                if int(start) < int(e.txt.cdsStart) or int(end) > int(e.txt.cdsEnd):
                    e.artefact = True
                    self.log.write(' '.join(("novel_intron artefact", e.contig, e.transcript, e.type, e.coordinate())) + "\n")


    def too_many_exons_skipped(self, match):
        "removes skipped exons if over 10 exons were skipped e.g. antibody gene"
        
        max_skipped = 10
        
        for e in match.events:
            if e.type == "skipped_exon" and len(e.exons) > max_skipped:
                e.artefact = True
                self.log.write(' '.join(("too-many-exons artefact", e.contig, e.transcript, e.type, e.coordinate())) + "\n")
                        
            
    def compare_model(self, align, txts):
        matches = []

        for txt in txts:
            result = self.match_exons(align.query, txt.full_name(), align.blocks, txt.exons, txt.chrom, strand=txt.strand)

            if not result.matched_blocks:
                continue

            result.model = txt.model
            result.model_weight = txt.weight
            result.txt = txt
            #result.align = align
            result.coord = "%s:%s-%s" % (align.target, align.blocks[0][0], align.blocks[-1][1])

            if result.events:
                for e in result.events:
                    e.txt = txt
                    e.align = align
                        
            #get coverage
            exon_bases = intspan.intersect(align.blocks, txt.exons)
            result.set_coverage(exon_bases, txt.length)

            if result.matched_blocks:
                matches.append(result)

        if matches:
            if len(matches) > 1:
                matches.sort(MatchResult.compare_matches)

            for m in matches:
                if m.events:
                    #set novelty
                    #novel events due to splitting of current events

                    novel_events_info = []
                    novel_events = {}

                    for i in range(len(m.events)):
                        e = m.events[i]
                        novel_events_info.extend(e.set_novelty(txts, matches))

                        for ne in novel_events_info:
                            variant = event.Event(ne['type'])
                            variant.align = align
                            tools.set_attrs(variant, ne)

                            if not novel_events.has_key(i):
                                novel_events[i] = []
                            novel_events[i].append(variant)

                    replaced_events = novel_events.keys()
                    replaced_events.sort(lambda x,y: y-x)

                    for i in replaced_events:
                        del m.events[i]
                    for i in replaced_events:
                        m.events.extend(novel_events[i])

                    #test for validity of events
                    self.as_artefacts(m)
                    self.deletion_artefact(m, align)
                    self.intron_artefact(m)
                    self.too_many_exons_skipped(m)

        else:
            result = MatchResult(align.query, None, len(align.blocks), 0, align)
            
            if len(align.blocks) >= 3:
                e = event.Event("novel_transcript")
                e.contig = align.query
                e.align = align
                e.chrom = align.target
                e.align_coords = align.blocks
                e.align_blocks = range(1,len(align.blocks)+1)
                result.events.append(e)
                
            matches.append(result)

        return matches
            

    def match_exons(self, align_id, txt_id, blocks, exons, chrom=None, strand=None):
        if self.log:
            self.log.write("%s %d:\t%s\n" % (align_id, len(blocks), pp.pformat(blocks).replace("\n", " ")))
            self.log.write("%s %d:\t%s\n" % (txt_id, len(exons), pp.pformat(exons).replace("\n", " ")))

        result = MatchResult(align_id, txt_id, len(blocks), len(exons))
        result.perfect_matches = 0

        #find start (first overlap of alignment block and exon) first
        b = 0
        e = 0
        b_start = None
        e_start = None

        #identify starting matching block 0 -> when they overlap
        while b < len(blocks):
            block = blocks[b]

            while e < len(exons):
                exon = exons[e]

                #if not overlap, on to next exon
                if int(block[0]) != int(exon[0]) and int(block[1]) != int(exon[1]):
                #if not intspan.overlap(block, exon):
                    e += 1
                    continue
                
                #last block (but not only block) overlap first exon, not real, skip
                elif b == len(blocks)-1 and e == 0 and len(blocks) > 1 and len(exons) > 1:
                    break

                #last exon overlap first block (but not only block), not real, skip
                elif e == len(exons)-1 and b == 0 and len(blocks) > 1 and len(exons) > 1:
                    break

                #start identified, done get out of exon loop
                b_start = b
                e_start = e
                break

            #start not identified, next alignment block
            if b_start == None and e_start == None:
                e = 0
                b += 1
            #start identified, done get out of block loop
            else:
                break

        if b_start >= 0 and e_start >= 0:
            b = b_start
            e = e_start

            if self.log:
                self.log.write("start %d %d %d %d\n" % (b, e, len(blocks), len(exons)))

            if b_start > 0:
                if e_start == 0:
                    last_novel_utr_exon = None
                    for bb in range(0, b_start):
                        if int(blocks[bb][1]) < int(exons[0][0]):
                            last_novel_utr_exon = bb 
                        else:
                            if self.log:
                                self.log.write("novel intron %d %d\n" % (bb, bb+1))

                            variant = event.Event("novel_intron")
                            tools.set_attrs(variant, {'contig':align_id, 'transcript':txt_id, 'chrom':chrom})
                            tools.set_attrs(variant, {'align_blocks':range(bb+1, bb+3), 'align_coords':blocks[bb:bb+2]})
                            tools.set_attrs(variant, {'exons':[1], 'exon_coords':[exons[0]]})
                            result.events.append(variant)

                    if last_novel_utr_exon != None:
                        if self.log:
                            self.log.write("novel 5utr %s %s %s-%s\n" % (align_id, txt_id, blocks[0][0], blocks[last_novel_utr_exon][1]))

                        variant = event.Event("novel_utr")
                        tools.set_attrs(variant, {'contig':align_id, 'transcript':txt_id, 'chrom':chrom})
                        tools.set_attrs(variant, {'align_blocks':range(1, last_novel_utr_exon+2), 'align_coords':blocks[0:last_novel_utr_exon+1]})
                        tools.set_attrs(variant, {'exons':[1], 'exon_coords':[exons[0]]})
                        if strand == '+':
                            variant.prime = '5'
                        else:
                            variant.prime = '3'
                        result.events.append(variant)
                    
                else:
                    for bb in range(1, b_start):
                        if self.log:
                            self.log.write("novel exon %d %d\n" % (bb,bb))

                        variant = event.Event("novel_exon")
                        tools.set_attrs(variant, {'contig':align_id, 'transcript':txt_id, 'chrom':chrom})
                        tools.set_attrs(variant, {'align_blocks':[bb+1], 'align_coords':[blocks[bb]]})
                        tools.set_attrs(variant, {'exons':[e_start], 'exon_coords':[exons[e_start]]})
                        result.events.append(variant)

            while b < len(blocks):
                block = blocks[b]
                        
                while e < len(exons):
                    exon = exons[e]

                    matched = self.match(blocks, exons, b, e)
                    if self.log:
                        self.log.write("%d %d %s %s %s\n" % (b, e, pp.pformat(block), pp.pformat(exon), matched))

                    #if matched:
                    #if matched and b != len(blocks)-1:
                    if matched and (not matched == 'sync' or b == len(blocks)-1):
                        result.matched_exons.append(e+1)
                        result.matched_blocks.append(b+1)

                        if matched == 'perfect':
                            result.perfect_matches += 1

                        if not matched == 'perfect' and len(blocks) > 1:
                            #not e == 0: not considering first exon (start 5'utr)
                            #not (b == 0 and int(block[0]) > int(exon[0])): first block and overshoot and overridden by last condition
                            #not (int(exon[0]) < int(blocks[b-1][-1])): start of exon not going over to previous block - not novel intron
                            #not (int(block[0]) < int(exons[e-1][-1])): start of block not going over to previous exon - not retained intron
                            #not b == 0: not considering first block (overrides 2nd condition -- temporarily turned off, because see a real extension of an internal exon
                            if matched == 'as5' and not e == 0 and not b == 0 and not (b == 0 and int(block[0]) > int(exon[0])) and not (b > 0 and int(exon[0]) < int(blocks[b-1][-1])) and not (e > 0 and int(block[0]) < int(exons[e-1][-1])):
                                if self.log:
                                    self.log.write("5as %s %s %s-%s\n" % (align_id, txt_id, block[0], block[1]))

                                if strand == '-':
                                    variant = event.Event("AS3")
                                else:
                                    variant = event.Event("AS5")
                                tools.set_attrs(variant, {'left':True, 'right':False})
                                tools.set_attrs(variant, {'contig':align_id, 'transcript':txt_id, 'chrom':chrom, 'edge':'left'})
                                tools.set_attrs(variant, {'align_blocks':[b+1], 'align_coords':blocks[b]})
                                tools.set_attrs(variant, {'exons':[e+1], 'exon_coords':exons[e]})
                                result.events.append(variant)

                            #not e == len(exons)-1: not considering last exon (end 3'utr)
                            #not (b == len(blocks)-1 and int(block[-1]) < int(exon[-1])): last block overshoot an internal exon - overridden by last condition
                            #not b == len(blocks)-1: not considering last block: overrides 2nd condition -- temporarily turned off
                            if matched == 'as3' and not e == len(exons)-1 and not b == len(blocks)-1:
                                if self.log:
                                    self.log.write("3as %s %s %s-%s\n" % (align_id, txt_id, block[0], block[1]))

                                if strand == '-':
                                    variant = event.Event("AS5")
                                else:
                                    variant = event.Event("AS3")
                                tools.set_attrs(variant, {'left':False, 'right':True})
                                tools.set_attrs(variant, {'contig':align_id, 'transcript':txt_id, 'chrom':chrom, 'edge':'right'})
                                tools.set_attrs(variant, {'align_blocks':[b+1], 'align_coords':blocks[b]})
                                tools.set_attrs(variant, {'exons':[e+1], 'exon_coords':exons[e]})
                                result.events.append(variant)

                            if matched == 'as53' and not e == 0 and not b == 0 and not (b == 0 and int(block[0]) > int(exon[0])) and not (b > 0 and int(exon[0]) < int(blocks[b-1][-1])) and not (e > 0 and int(block[0]) < int(exons[e-1][-1])) and not e == len(exons)-1 and not b == len(blocks)-1:
                                if self.log:
                                    self.log.write("as53 %d %d\n" % (b, b+1))

                                variant = event.Event("AS53")
                                tools.set_attrs(variant, {'contig':align_id, 'transcript':txt_id, 'chrom':chrom})
                                tools.set_attrs(variant, {'align_blocks':[b+1], 'align_coords':[blocks[b]]})
                                tools.set_attrs(variant, {'exons':[e+1], 'exon_coords':[exons[e]]})
                                result.events.append(variant)

                        #record matches = CHECK THIS!!!!!!!!!!
                        if e == 0 or e == len(exons)-1:
                            if matched == 'perfect':
                                score = 2
                            #'as53' will have no score
                            elif matched == 'as5' or matched == 'as3':
                                score = 1
                            else:
                                score = 0

                            if e == 0:
                                result.first = score

                            else:
                                result.last = score

                        else:
                            if matched == 'perfect':
                                result.middle += 2

                            #start(5') is different, end(3') is the same
                            elif matched == 'as5' and b == 0:
                                result.middle += 1

                            #start(5') is the same, end(3') is different
                            elif matched == 'as3' and b == len(blocks)-1:
                                result.middle += 1

                        b += 1
                        e += 1
                        break

                    #novel exon
                    elif int(block[1]) < int(exon[0]):
                        #increment block, until finding matching block/exon
                        bb = b + 1

                        matched = False
                        while not matched and bb <= len(blocks)-1:
                            matched = self.match(blocks, exons, bb, e)
                            if not matched:
                                bb += 1

                        #if bb == len(blocks)-1, check if loop stops because of match or last index
                        #if loops stops because of finding matching exon, then novel exons is between b and bb
                        if matched:
                            if self.log:
                                self.log.write("novel exons %d %d\n" % (b, bb))

                            variant = event.Event("novel_exon")
                            tools.set_attrs(variant, {'contig':align_id, 'transcript':txt_id, 'chrom':chrom})
                            tools.set_attrs(variant, {'align_blocks':range(b+1,bb+1), 'align_coords':blocks[b:bb]})
                            tools.set_attrs(variant, {'exons':[e,e+1], 'exon_coords':exons[e-1:e+1]})
                            result.events.append(variant)
 
                        #set next comparison to be bb, if it's the last index, then handle it outside??, it will handle alt.splicing
                        b = bb
                        break

                    #skipped exon
                    elif int(block[0]) > int(exon[1]):
                        #increment exon, until finding matching block/exon
                        ee = e + 1
                        
                        matched = False
                        while not matched and ee <= len(exons)-1:
                            matched = self.match(blocks, exons, b, ee)
                            if not matched:
                                ee += 1

                        #requires ee to be matched
                        if matched:                       
                            if self.log:
                                self.log.write("skipped exons %d %d\n" % (e, ee))

                            variant = event.Event("skipped_exon")
                            tools.set_attrs(variant, {'contig':align_id, 'transcript':txt_id, 'chrom':chrom})
                            #tools.set_attrs(variant, {'align_blocks':[b], 'align_coords':[blocks[b-1]]})
                            tools.set_attrs(variant, {'align_blocks':range(b,b+2), 'align_coords':blocks[b-1:b+1]})
                            tools.set_attrs(variant, {'exons':range(e+1,ee+1), 'exon_coords':exons[e:ee]})
                            result.events.append(variant)

                        e = ee
                        break

                    #retained intron
                    elif int(block[1]) > int(exon[1]):
                        #increment exon, until finding matching block/exon
                        ee = e + 1

                        if ee <= len(exons)-1 and int(exons[ee][0]) < int(block[1]):
                            matched = False

                            while (not matched or matched == 'sync') and ee <= len(exons)-1:
                                matched = self.match(blocks, exons, b, ee)
                                if not matched or matched == 'sync':
                                    ee += 1

                            if matched and not matched == "as3":
                                if self.log:
                                    self.log.write("retained intron %d %d\n" % (e, ee))

                                variant = event.Event("retained_intron")
                                tools.set_attrs(variant, {'contig':align_id, 'transcript':txt_id, 'chrom':chrom})
                                tools.set_attrs(variant, {'align_blocks':[b+1], 'align_coords':[blocks[b]]})
                                tools.set_attrs(variant, {'exons':range(e+1,ee+2), 'exon_coords':exons[e:ee+1]})
                                result.events.append(variant)
                        else:
                            b = b + 1

                        e = ee
                        break

                    #novel intron/indel
                    elif int(block[1]) < int(exon[1]):
                        #increment block, until finding matching block/exon
                        bb = b + 1

                        if bb <= len(blocks)-1 and int(blocks[bb][0]) < int(exon[1]):
                            matched = False

                            while not matched and bb <= len(blocks)-1:
                                #matched = self.match(blocks, exons, bb, e, move="block")
                                matched = self.match(blocks, exons, bb, e)
                                if not matched:
                                    bb += 1

                            if matched:
                                if self.log:
                                    self.log.write("novel intron %d %d\n" % (b, bb))

                                variant = event.Event("novel_intron")
                                tools.set_attrs(variant, {'contig':align_id, 'transcript':txt_id, 'chrom':chrom})
                                tools.set_attrs(variant, {'align_blocks':range(b+1,bb+2), 'align_coords':blocks[b:bb+1]})
                                tools.set_attrs(variant, {'exons':[e+1], 'exon_coords':[exons[e]]})
                                result.events.append(variant)
                        else:
                            e = e + 1
                            
                        b = bb
                        break

                    else:
                        if self.log:
                            self.log.write("??\n")
                        e += 1
                        break

                    break
                
                #make sure the loop ends
                if e > len(exons)-1:
                    break
                
            #if last matched/synced is not last block and is last exon
            #print b, blocks[b], exons[e-1]
            if b <= len(blocks)-1:
                for bb in range(b, len(blocks)):
                    if int(blocks[bb][0]) > int(exons[e-1][1]):
                        if self.log:
                            self.log.write("novel 3utr %s %s %d %d %s-%s\n" % (align_id, txt_id, bb, e, exons[0][0], exons[-1][-1]))
                        variant = event.Event("novel_utr")
                        tools.set_attrs(variant, {'contig':align_id, 'transcript':txt_id, 'chrom':chrom})
                        tools.set_attrs(variant, {'align_blocks':range(bb+1,len(blocks)+1), 'align_coords':blocks[bb:]})
                        tools.set_attrs(variant, {'exons':[len(exons)], 'exon_coords':[exons[-1]]})
                        if strand == '-':
                            variant.prime = '5'
                        else:
                            variant.prime = '3'
                        result.events.append(variant)
                        break
                    
                    else:
                        if self.log:
                            self.log.write("novel intron %d %d\n" % (bb-1, bb))

                            variant = event.Event("novel_intron")
                            tools.set_attrs(variant, {'contig':align_id, 'transcript':txt_id, 'chrom':chrom})
                            tools.set_attrs(variant, {'align_blocks':range(bb, bb+2), 'align_coords':blocks[bb-1:bb+1]})
                            tools.set_attrs(variant, {'exons':[e], 'exon_coords':[exons[e-1]]})
                            result.events.append(variant)

        else:
            if self.log:
                self.log.write("do not overlap could not find start")

        if self.log:
            self.log.write("----\n")

        #only one matching block and there is events => fishy
        if len(result.matched_blocks) == 1 and result.perfect_matches == 0 and len(blocks) > 1:
            del result.matched_blocks[:]
            del result.matched_exons[:]

        if result.events and not result.matched_blocks:
            del result.events[:]
            
        return result


    def match(self, blocks, exons, b, e, move=None):
        match = False

        if int(blocks[b][0]) == int(exons[e][0]) and int(blocks[b][1]) == int(exons[e][1]):
            match = "perfect"

        #single block and embedded within intron
        elif len(blocks) == 1:
            #one block embedded within an exon
            if intspan.subsume(blocks[b], exons[e]):
                match = "perfect"
            elif int(blocks[b][1]) > int(exons[e][1]):
                match = "as3"
            elif not e == 0 and int(blocks[b][0]) < int(exons[e][0]):
                match = "as5"

        elif len(exons) == 1:
            if intspan.subsume(blocks[b], exons[e]):
                match = "sync"

        elif e == len(exons)-1:
            #if not overlap, or cut into next exon return false
            if intspan.overlap(blocks[b], exons[e]):
                #last block, last exon
                if b == len(blocks)-1:
                    if int(blocks[b][0]) == int(exons[e][0]):
                        match = "perfect"
                    else:
                        match = "as5"
                #not last block, novel exon
                else:
                    #2nd condition check if it has retained intron
                    if int(blocks[b][0]) == int(exons[e][0]) or int(blocks[b][0]) < int(exons[e-1][1]):
                        match = "as3"
		    else:
			match = "as5"
                    #elif int(blocks[b][1]) == int(exons[e][1]):
                        #match = "as5"
                    #elif int(blocks[b+1][0]) > int(exons[e][1]):
                        #match = "as53"
                    #else:
                        #match = "sync"

        elif b == len(blocks)-1:
            #if not overlap, or cut into next exon return false
            #first check overlap
            if intspan.overlap(blocks[b], exons[e]):
                #last block vs middle exon
                if e < len(exons)-1:
                    #not retained intron
                    if int(blocks[b][1]) < int(exons[e+1][0]):
                        if int(blocks[b][0]) == int(exons[e][0]) or int(blocks[b][0]) < int(exons[e-1][1]):
                            match = "as3"
                        elif int(blocks[b][-1]) == int(exons[e][-1]):
                            match = "as5"
                        else:
                            match = "as53"
                    #retained intron
                    else:
                        match = "sync1"
                #last block vs last exon (last exon = only check as5)
                elif not (int(blocks[b][0]) == int(exons[e][0])):
                    match = "as5"
                #perfect = 3' end not match
                else:
                    match = "perfect"
        
        elif intspan.subsume(blocks[b], exons[e]) or intspan.subsume(exons[e], blocks[b]):
            if int(exons[e][1]) < int(blocks[b+1][0]) and int(blocks[b][1]) < int(exons[e+1][0]):
                if int(blocks[b][0]) == int(exons[e][0]):
                    match = "as3"
                elif int(blocks[b][1]) == int(exons[e][1]):
                    match = "as5"
                else:
                    if int(blocks[b+1][0]) > int(exons[e][1]):
                        match = "as53"
                    else:
                        match = "sync"
            else:
                match = "sync"

        #as5 or as3
        elif int(blocks[b][1]) < int(exons[e+1][0]) and int(exons[e][1]) < int(blocks[b+1][0]):
            if int(blocks[b][0]) == int(exons[e][0]):
                match = "as3"
            elif int(blocks[b][1]) == int(exons[e][1]):
                match = "as5"
            elif intspan.overlap(blocks[b], exons[e]):
                match = "as53"
            else:
                match = "sync"
                    
        elif int(blocks[b][1]) < int(exons[e+1][0]) and int(exons[e][1]) < int(blocks[b+1][0]) and int(exons[e+1][0]) < int(blocks[b+1][1]) and int(blocks[b+1][0]) < int(exons[e+1][1]):
            if int(blocks[b][0]) == int(exons[e][0]):
                match = "as3"
            elif int(blocks[b][1]) == int(exons[e][1]):
                match = "as5"
            else:
                match = "as53"
                
        #move on
        elif int(blocks[b][0]) > int(exons[e][1]) and int(blocks[b][1]) < int(exons[e+1][0]):
            match = "sync"

##        elif move == "block" and int(blocks[b][1]) > int(exons[e][1]):
##            print "ee",blocks[b],exons[e]
##            match = "sync"

        #elif int(blocks[b][1]) > int(exons[e][1]) and int(blocks[b][0]) > int(exons[e][0]):
            #print "ee",blocks,exons,b,e
        #    match = "sync"
            
        else:
            pass

        return match

    def output_coverage(self, contig2txts, txt_objs, outfile):
        out = open(outfile, 'w')

        headers = '\t'.join(('transcript', 'gene', 'total_coverage', 'transcript_length', 'best_contig', 'best_contig_coverage', 'nbr_contigs', 'contigs', 'coverage'))

        out.write(headers + "\n")
        
	for model in self.model_order:
	    if self.coverage_model and model != self.coverage_model:
		continue
	    
	    if not contig2txts.has_key(model):
		continue
	    
            all_blocks = []
            for txt_id in contig2txts[model].keys():
                del all_blocks[:]
                txt = txt_objs[model][txt_id]
                gene = 'na'
                if txt.alias:
                    gene = txt.alias
                contigs = contig2txts[model][txt_id].keys()

                best_align, best_contig, best_coverage = None, None, None
                for contig in contigs:
                    align, coverage = contig2txts[model][txt_id][contig]
                    blocks_copy = align.blocks[:]
                    all_blocks.append(blocks_copy)

                    if not best_contig or coverage > best_coverage:
                        best_align, best_contig, best_coverage = align, contig, coverage
                        
                union_blocks = intspan.union(all_blocks)
                coverage_sum = intspan.intersect(txt.exons, union_blocks)
                coverage = float(coverage_sum)/float(txt.length)

                #txt gene coverage_sum txt_len best-contig best-contig-coverage num_contigs list_contigs coverage
                line = "\t".join((txt_id, gene, str(coverage_sum), str(txt.length), best_contig, "%.3f" % (best_coverage), str(len(contigs)), ','.join(contigs), "%.3f" % (coverage)))
                
                out.write(line + "\n")

        out.close()
    

    def output_mapping(self, output_file):
        "output mapping to all models"
        
        ff = open(output_file, 'w')
        for contig, matches in self.full_matches.iteritems():
            for match in matches:
                ff.write(match.report())

        for matches in self.partial_matches.values():
            for match in matches:
                ff.write(match.report())

        for contig, matches in self.non_matches.iteritems():
            ff.write(contig + " no_match " + str(len(matches[0].align.blocks)) + "\n")
            
        ff.close()
                    
    def group_events(self, events):
        grouped_events = {}
        for e in events:
            if e.novel and not e.artefact and e.type != 'novel_transcript':
                coord = e.coordinate()
                if not grouped_events.has_key(e.type):
                    grouped_events[e.type] = {coord:[e]}
                else:
                    if not grouped_events[e.type].has_key(coord):
                        grouped_events[e.type][coord] = [e]
                    else:
                        grouped_events[e.type][coord].append(e)

        self.group_novel_events([e for e in events if e.type == 'novel_transcript'], grouped_events)

        return grouped_events

    def group_novel_events(self, events, grouped_events):
        #group novel transcripts
        novel_transcripts = {}
        count = 0

        for e in events:
            chrom,start,end = re.split('[:-]', e.coordinate())
                                         
            if not novel_transcripts.has_key(chrom):
                novel_transcripts[chrom] = {}
                novel_transcripts[chrom][start + "-" + end] = [e]
            else:
                found = False
                for start_end in sorted(novel_transcripts[chrom].keys()):
                    ee = novel_transcripts[chrom][start_end][0]

                    if intspan.same_blocks(e.align_coords, ee.align_coords) or intspan.same_blocks(ee.align_coords, e.align_coords):
                        novel_transcripts[chrom][start_end].append(e)
                        found = True
                        break

                if not found:
                    if not novel_transcripts[chrom].has_key(start + "-" + end):
                        novel_transcripts[chrom][start + "-" + end] = [e]
                    else:
                        novel_transcripts[chrom][start + "-" + end + "." + str(count)] = [e]

            count += 1

        type = 'novel_transcript'
 
        if len(novel_transcripts.keys()) > 0:
            grouped_events[type] = {}
        
        for chrom in novel_transcripts.keys():
            for start_end in novel_transcripts[chrom].keys():
                coord = chrom + ":" + start_end
                
                if not grouped_events[type].has_key(coord):
                    grouped_events[type][coord] = []
                
                for e in novel_transcripts[chrom][start_end]:
                    grouped_events[type][coord].append(e)


    def output_events(self, grouped_events, output_file, summary_file, bed_file=None):
        output = open(output_file, 'w')
        event_counts = {}

        headers = event.Event.headers[:]
        headers.insert(0, 'id')
        if self.contigs_bam:
            headers.extend(event.Event.headers_support)
        output.write('\t'.join(headers) + "\n")

        count1 = 1
        for type in sorted(grouped_events):
            event_counts[type] = 0
            
            for coord in grouped_events[type].keys():
                event_counts[type] += 1
                
                count2 = 1
                for e in grouped_events[type][coord]:
                    if len(grouped_events[type][coord]) == 1:
                        count = count1
                    else:
                        count = "%d.%d" % (count1, count2)

                    if self.debug:
                        self.log.write("outputting event %s for contig %s\n" % (e.type, e.contig))

                    output.write(str(count) + "\t" + e.output() + "\n")

                    count2 += 1

                count1 += 1

        output.close()

        summary = open(summary_file, 'w')
        for type in sorted(event_counts.keys()):
            summary.write(type + '\t' + str(event_counts[type]) + '\n')
        summary.close()

        if bed_file:
            self.output_bed(grouped_events, bed_file)

    def output_bed(self, grouped_events, output_file):
        types = ['novel_exon', 'AS53', 'skipped_exon', 'retained_intron', 'novel_utr', 'novel_intron', 'novel_transcript', 'AS5', 'AS3']
        priority_base = 0.00001
        priority_increment = 0.000001
        colors = ['178,34,34', '178,34,34', '30,144,255', '0,128,0', '105,105,105', '72,61,139', '128,0,128', '255,140,0',  '255,140,0']

        output = open(output_file, 'w')
        for i in range(len(types)):
            type = types[i]
            color = colors[i]
            priority = priority_base + i * priority_increment

            if grouped_events.has_key(type):
                output.write("track name=%s description=%s priority=%.6f color=%s visibility=pack\n" % (type, type, priority, color))
            
                for coord in grouped_events[type]:
                    for e in grouped_events[type][coord]:
                        chrom, start, end = re.split('[:-]', e.coordinate())
                        line = "\t".join([chrom, str(int(start)-1), end, e.contig])
                        output.write(line + "\n")
                    
        output.close()
		

if __name__ == '__main__':
    usage = "Usage: %prog alignment_file genome"

    parser = OptionParser(usage=usage, version="%prog " + VERSION)
    parser.add_option("-m", "--models", dest="models", help="model order e.g. k,e,r,a, if not specified, model order specified in config file will be used")
    parser.add_option("-l", "--log", dest="log", action="store_true", default=False)
    parser.add_option("-d", "--debug", dest="debug", action="store_true", default=False)
    parser.add_option("-o", "--outdir", dest="outdir", help="output directory")
    parser.add_option("-f", "--sequence", dest="fasta", help="fasta file")
    parser.add_option("-k", "--kmer", dest="k", help="k-mer length")
    parser.add_option("-r", "--ref", dest="ref", help="use reference sequence for deducing splice sites for psl", action="store_true", default=False)
    parser.add_option("-C", "--contigs_bam", dest="contigs_bam", help="reads to contigs bam file")
    parser.add_option("-B", "--genome_bam", dest="genome_bam", help="reads to genome bam file")
    parser.add_option("-L", "--library", dest="lib", help="library name for parsing contigs bam file")
    parser.add_option("-S", "--spanning_reads", dest="spanning_reads", help="minimum number of spanning reads. Default=2", type='int', default=2)
    parser.add_option("-R", "--coverage_diff", dest="coverage_diff", help="maximum coverage diff(ratio) between event block and neighbors. Default=100", type='int', default=100)
    parser.add_option("-s", "--add_contig_reads", dest="add_contig_reads", action="store_true", default=False)
    parser.add_option("-M", "--coverage_model", dest="coverage_model", help="single model used for reporting coverage e.g. e")
    
    (options, args) = parser.parse_args()
    tools.output_log(VERSION, sys.argv, [vars(options)], options.outdir)
    
    if len(args) == 2:
        models = None
        if options.models:
            models = options.models.split(',')

        if options.outdir:
            if not os.path.isdir(options.outdir):
                os.mkdir(options.outdir)

        if options.log:
            log_file = options.outdir + "/log.txt"

        refseq = None
        if options.ref:
	    refseq = tools.get_refseq_from_2bit(args[1])

        #do this before extracting alignments because need to get splice file info
        mm = ModelMatcher(genome=args[1], debug=options.debug, outdir=options.outdir, model_order=models, contigs_bam=options.contigs_bam, genome_bam=options.genome_bam, lib=options.lib, min_spanning_reads=options.spanning_reads, max_coverage_diff=options.coverage_diff, refseq=refseq)
	
	mm.coverage_model = options.coverage_model
	mm.add_contig_reads = options.add_contig_reads
	
        #get alignments from track
        sys.stderr.write("get alignments...")
        trk = track.Track(args[0])

        ext = os.path.splitext(args[0])[1]

        sys.stderr.write("extracting alignments...")
	
        if ext == ".gff":
            mm.aligns = trk.parse_gff()
        elif ext == ".psl":
            mm.aligns = trk.parse_psl(minimum=True, refseq=refseq, splice_motif_file=mm.splice_file)
        sys.stderr.write("done\n");

        #get contig sequences
        contigs = None
        ass = None

        if options.fasta:
            sys.stderr.write("getting contig sequences...")
            contigs_with_aligns = [a.query for a in mm.aligns]
            mm.assembly = assembly.Assembly(None, fasta=options.fasta)
            contigs = mm.assembly.get_contigs(ids=contigs_with_aligns, sequence=True)
            sys.stderr.write("done\n")

            #linking alignment to contig
            if contigs:
                contig_dict = dict((c.num, c) for c in contigs)

                for align in mm.aligns:
                    if contig_dict.has_key(align.query):
                        align.contig = contig_dict[align.query]

                        if options.k:
                            align.contig.k = options.k
                        elif "k" in align.contig.num:
                            align.contig.k = align.contig.num.split(':')[0].lstrip("k")

        #funnel through transcript models
        if mm.model_order:
            mm.match_models()
            
    else:
        parser.error("incorrect number of arguments")
