import os, re
import annotations
from utilities.tools import is_integer
import ConfigParser
from utilities.overlap_coord import OverlapCoord

PACKAGE_DIR = "/".join(os.path.abspath(__file__).split("/")[:-2])

class Transcript:

    name = None
    chrom = None
    strand = None
    txStart = txEnd = None
    cdsStart = cdsEnd = None
    exonCount = None
    exons = None
    gene = None
    alias = None
    model = None
    weight = None
    length = 0

    def __init__(self, name):
        self.name = name

    def full_name(self):
        alias = "NA"
        if self.alias:
            alias = self.alias
        
        return "%s(%s)" % (self.name, alias)

    def details(self):
        return "%s %s %s %s-%s %s" % (self.full_name(), self.chrom, self.strand, self.txStart, self.txEnd, self.exonCount)

    def cds_length(self):
        if self.cdsStart and self.cdsEnd and is_integer(self.cdsStart) and is_integer(self.cdsEnd):
            return int(self.cdsEnd) - int(self.cdsStart)

        return 0
    
    def coding_type(self):
        if self.cdsStart == None or self.cdsEnd == None:
            return 'NA'
        elif is_integer(self.cdsStart) and is_integer(self.cdsEnd) and self.cdsStart != self.cdsEnd:
            return 'CODING'
        else:
            return 'NONCODING'
        
    def utr(self, end):
        """
        Return UTR coordinate in [start,end] 
        User must specify '5' or '3'
        If noncoding, will return None
        """
        utr3 = utr5 = None
        if self.coding_type() == 'CODING':
            if self.strand == '+':
                utr5 = [int(self.txStart), int(self.cdsStart) - 1]
                utr3 = [int(self.cdsEnd) + 1, int(self.txEnd)]
            else:
                utr3 = [int(self.txStart) + 1, int(self.cdsStart)]
                utr5 = [int(self.cdsEnd) + 1, int(self.txEnd)]
            
            if end == 5:
                return utr5
            elif end == 3:
                return utr3
            else:
                return None
        else:
            return utr3
            
    @classmethod
    def find_by_name(cls, model, annot_file, txt_name):
        txts = []

        if os.path.exists(annot_file):
            for line in open(annot_file, 'r'):
                if txt_name in line:
                    txt = {
                        'e': annotations.ensembl.parse_line,
                        'r': annotations.refGene.parse_line,
                        'k': annotations.knownGene.parse_line,
                        'a': annotations.aceview.parse_line,
                        'x': annotations.ensg.parse_line
                        }[model](line)

                    if txt.name == txt_name or txt.alias == txt_name:
                        txts.append(txt)

        return txts
    
    @classmethod
    def prepare_overlap(cls, genome, models):
        overlaps = {}
    
        config = ConfigParser.ConfigParser()
        config.read(PACKAGE_DIR + "/configs/model_matcher.cfg")
    
        for m in models.split(','):       
            config = ConfigParser.ConfigParser()
            config.read(PACKAGE_DIR + "/configs/model_matcher.cfg")
            annot_file = index = None

            if genome in config.sections():
                for field in config.options(genome):
                    if field.lower() == m.lower():
                        annot_file = PACKAGE_DIR + "/annotations/" + genome + "/" + config.get(genome, field)

                        if annot_file and os.path.exists(annot_file):
                            index_file = annot_file.replace('.txt', '.idx')
                        
                            txt_overlap = OverlapCoord(annot_file, index_file)
                            txt_overlap.extract_index()
                        
                            overlaps[m] = txt_overlap
                        
        return overlaps

    @classmethod
    def find_overlaps(cls, overlaps, models, chrom, start, end, only_coding=False):
        txts = []
        for model in models:
            lines = overlaps[model].overlap(chrom, int(start), int(end))
            for line in lines:
                txt = {
                    'e': annotations.ensembl.parse_line,
                    'r': annotations.refGene.parse_line,
                    'k': annotations.knownGene.parse_line,
                    'a': annotations.aceview.parse_line,
                    'x': annotations.ensg.parse_line
                    }[model](line)
                
                if only_coding and txt.coding_type().upper() != 'CODING':
		    continue
                
                txts.append(txt)
                    
        return txts